! module for 2D forward modeling using a time-domain FD method
!
module a2d

use global
implicit none

integer, private              :: ix, iz, it, is, is1, is2, isx, isz, ig, igx, igz, &
                                 i, j, jj
integer, private, allocatable :: it_mute(:)
real,    private              :: alpha, beta, a, d1, d2, kappa, obs_fac, calc_fac
real,    private, allocatable :: boundary(:,:), p_end(:,:), p_end_minus1(:,:), &
                                 u_end(:,:), w_end(:,:), p0(:,:), p1(:,:), p(:,:), dp(:,:),&
                                 csg_calc(:,:), csg_obs(:,:), u(:,:), w(:,:), dwin(:,:)

contains

!-----------------------------------------------------------------------------------------
subroutine modeling_mpi( method, message, par, coord, s, c, den, fs, nx_pml, nz_pml, npml, &
                         damp, compute_residual, res )
use datatype
use mmi_mpi

type(param),       intent(in)    :: par
type(acquisition), intent(in)    :: coord
character(*),      intent(in)    :: method
logical,           intent(in)    :: message, compute_residual
integer,           intent(in)    :: fs(:), nx_pml, nz_pml, npml
real,              intent(in)    :: damp(:,:), c(:,:), den(:,:)
real,              intent(inout) :: s(:)
real,              intent(out)   :: res
integer                          :: i, j
real                             :: res_total

    if (method(1:9) == 'staggered') then
        if (.not.allocated(boundary)) allocate(boundary(par%nx*9+par%nz*6,par%nt))
        if (.not.allocated(p_end))    allocate(p_end(par%nz,par%nx))
        if (.not.allocated(u_end))    allocate(u_end(par%nz,par%nx))
        if (.not.allocated(w_end))    allocate(w_end(par%nz,par%nx))
    else
      !  if (.not.allocated(boundary))     allocate(boundary(par%nx*5+par%nz*4,par%nt))
        if (.not.allocated(boundary))     allocate(boundary(par%nx*7+par%nz*6,par%nt))
        if (.not.allocated(p_end))        allocate(p_end(par%nz,par%nx))
        if (.not.allocated(p_end_minus1)) allocate(p_end_minus1(par%nz,par%nx))
    endif
    
    ! Initialize residual
    res = 0.0
    call get_assigned(1, coord%ns, is1, is2)
    
    do is = is1, is2, par%skipshot+1
        if (message) then
          write(*,*) 'Process ', rank, ', shot', is
          flush(6)
        endif
    
        if (method(1:9) == 'staggered') then
            call staggered42_modeling_is(is, par, coord, s, c, den, fs, nx_pml, nz_pml, npml, damp, &
                    message, boundary, p_end, u_end, w_end, compute_residual, res)
        else
            call a2d42_modeling_is(is, par, coord, s, c, fs, nx_pml, nz_pml, npml, damp, &
                    message, boundary, p_end, p_end_minus1, compute_residual, res)
        endif
    enddo
    
    if (compute_residual) then
        call MPI_Allreduce(res,res_total,1, MPI_REAL,MPI_SUM,MPI_COMM_WORLD,ierr)
        res = sqrt(res_total)
    endif

end subroutine modeling_mpi

!-----------------------------------------------------------------------------------------!
! Solving the second-order constant-density acoustic wave equation using a regular-grid
! explicit finite-difference method.

subroutine a2d42_modeling_is(is, par, coord, s, c, fs, nx_pml, nz_pml, npml, damp, &
                  store_boundary, boundary, p_end, p_end_minus1, compute_residual, res)

use io
use datatype
use math
use modeling
use source
use string

type(param),       intent(in)    :: par
type(acquisition), intent(in)    :: coord
logical,           intent(in)    :: store_boundary, compute_residual
integer,           intent(in)    :: is, fs(:), nx_pml, nz_pml, npml
real,              intent(in)    :: s(:), damp(:,:), c(:,:)
real,              intent(out)   :: boundary(:,:), p_end(:,:), p_end_minus1(:,:)
real,              intent(inout) :: res
real                             :: temp_res
integer                          :: it1,it2 ! adding window to traces

! Memory allocations
allocate(p0(nz_pml,nx_pml))
allocate(p1(nz_pml,nx_pml))
allocate(p(nz_pml,nx_pml))
allocate(csg_calc(par%nt,coord%ngmax))
allocate (dwin(coord%ns, 2*coord%ngmax))

    ! Initialize arrays
    p0 = 0.0
    p1 = 0.0
    p  = 0.0
    csg_calc = 0.0
    
    ! Setup source and receiver position
    isx = npml+int(coord%xs(is)/par%dx)+1
    isz = npml+int(coord%zs(is)/par%dx)+1
    ! Source position is shifted downward 1 node if it is on the free surface.
    if (isz == fs(isx)) isz = isz + 1
    
    ! Add source to the initial pressure field
    p0(isz,isx) = s(1)
    p1(isz,isx) = s(2)
    
    do ig=1,coord%ng(is),1+par%skiptrace
        igx = npml + int(coord%xg(is,ig)/par%dx)+1
        igz = npml + int(coord%zg(is,ig)/par%dx)+1
        ! Geophone position is shifted downward 1 node if it is on the free surface.
        if (igz == fs(igx)) igz = igz + 1
        csg_calc(1,ig) = p0(igz,igx)
        csg_calc(2,ig) = p1(igz,igx)
    enddo
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    if (store_boundary) then
        jj = 0
        ! Store the 3 bottom lines
        do ix=npml+1,nx_pml-npml
            boundary(jj+1,1) = p0(nz_pml-npml  ,ix)
            boundary(jj+2,1) = p0(nz_pml-npml-1,ix)
            boundary(jj+3,1) = p0(nz_pml-npml-2,ix)
            jj = jj+2+1
        enddo
        ! Store the 3 left and 3 right lines
        do iz=npml+1,nz_pml-npml
            boundary(jj+1,1) = p0(iz,npml+1)
            boundary(jj+2,1) = p0(iz,npml+2)
            boundary(jj+3,1) = p0(iz,npml+3)
    
            boundary(jj+4,1) = p0(iz,nx_pml-npml)
            boundary(jj+5,1) = p0(iz,nx_pml-npml-1)
            boundary(jj+6,1) = p0(iz,nx_pml-npml-2)
            jj = jj+4+2
        enddo
        ! Store the 4 top lines
        do j=npml+1,nx_pml-npml
            boundary(jj+1,1) = p0(fs(j)  ,j)
            boundary(jj+2,1) = p0(fs(j)+1,j)
            boundary(jj+3,1) = p0(fs(j)+2,j)
            boundary(jj+4,1) = p0(fs(j)+3,j)
            jj = jj+3+1
        enddo
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        jj = 0
        ! Store the 3 bottom lines
        do ix=npml+1,nx_pml-npml
            boundary(jj+1,2) = p1(nz_pml-npml  ,ix)
            boundary(jj+2,2) = p1(nz_pml-npml-1,ix)
            boundary(jj+3,2) = p1(nz_pml-npml-2,ix)      
            jj = jj+2+1
        enddo
        ! Store the 3 left and 3 right lines
        do iz=npml+1,nz_pml-npml
            boundary(jj+1,2) = p1(iz,npml+1)
            boundary(jj+2,2) = p1(iz,npml+2)
            boundary(jj+3,2) = p1(iz,npml+3)
    
            boundary(jj+4,2) = p1(iz,nx_pml-npml)
            boundary(jj+5,2) = p1(iz,nx_pml-npml-1)
            boundary(jj+6,2) = p1(iz,nx_pml-npml-2)
            jj = jj+4+2
        enddo
        ! Store the 4 top lines
        do j=npml+1,nx_pml-npml
            boundary( jj+1, 2) = p1( fs(j)  , j )
            boundary( jj+2, 2) = p1( fs(j)+1, j )
            boundary( jj+3, 2) = p1( fs(j)+2, j )
            boundary( jj+4, 2) = p1( fs(j)+3, j )
            jj = jj+3+1
        enddo
    endif
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! Time marching loop
    do it=3,par%nt
    ! Update the pressure field
        !do ix=3,nx_pml-2
        !  do iz=fs(ix)+1,nz_pml-2
        do ix = 4, nx_pml-3
            do iz = fs(ix)+1, nz_pml-3
              alpha = c(iz,ix)*par%dt/par%dx
              alpha = alpha*alpha
              kappa = damp(iz,ix)*par%dt
    !          p(iz,ix) = (2.0+2.0*c1_2nd_order*alpha-kappa)*p1(iz,ix) - &
    !                     (1.0-kappa)*p0(iz,ix) + alpha* &
    !                     (c2_2nd_order*(p1(iz,ix+1)+p1(iz,ix-1)+p1(iz+1,ix)+p1(iz-1,ix)) + &
    !                      c3_2nd_order*(p1(iz,ix+2)+p1(iz,ix-2)+p1(iz+2,ix)+p1(iz-2,ix)))
              p(iz,ix) = (2.0-kappa)*p1(iz,ix) - (1.0-kappa)*p0(iz,ix) + &
                           alpha*prefact*( 2.0*c1_6th*p1(iz,ix) + &
                           c2_6th*( p1(iz,ix+1)+p1(iz,ix-1)+p1(iz+1,ix)+p1(iz-1,ix) ) + &
                           c3_6th*( p1(iz,ix+2)+p1(iz,ix-2)+p1(iz+2,ix)+p1(iz-2,ix) ) + &
                           c4_6th*( p1(iz,ix+3)+p1(iz,ix-3)+p1(iz+3,ix)+p1(iz-3,ix) ) )
            enddo
        enddo

        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        ! Add source
        beta = c(isz,isx)*par%dt
        beta = beta*beta
        p(isz,isx) = p(isz,isx) + beta*s(it)
    
        ! Free surface
        do ix=1,nx_pml
            p(fs(ix)-1,ix) = -p(fs(ix)+1,ix)
            p(fs(ix)-2,ix) = -p(fs(ix)+2,ix); !depends on the FD scheme
        enddo
    
        ! Output pressure seismogram
        do ig=1,coord%ng(is),1+par%skiptrace
            igx = npml + int(coord%xg(is,ig)/par%dx)+1
            igz = npml + int(coord%zg(is,ig)/par%dx)+1
            if (igz == fs(igx)) igz = igz + 1
            csg_calc(it,ig) = p(igz,igx)
        enddo
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        if (store_boundary) then
            jj = 0
            ! Store the 3 bottom lines
            do ix=npml+1,nx_pml-npml
                boundary(jj+1, it) = p(nz_pml-npml  , ix)
                boundary(jj+2, it) = p(nz_pml-npml-1, ix)
                boundary(jj+3, it) = p(nz_pml-npml-2, ix)
                jj = jj+2+1
            enddo
            ! Store the 3 left and 3 right lines
            do iz=npml+1,nz_pml-npml
                boundary(jj+1, it) = p(iz, npml+1)
                boundary(jj+2, it) = p(iz, npml+2)
                boundary(jj+3, it) = p(iz, npml+3)
              
                boundary(jj+4, it) = p(iz, nx_pml-npml)
                boundary(jj+5, it) = p(iz, nx_pml-npml-1)
                boundary(jj+6, it) = p(iz, nx_pml-npml-2)
                jj = jj+4+2
            enddo
            ! Store the 4 top lines
            do j=npml+1,nx_pml-npml
                boundary(jj+1, it) = p(fs(j)  , j)
                boundary(jj+2, it) = p(fs(j)+1, j)
                boundary(jj+3, it) = p(fs(j)+2, j)
                boundary(jj+4, it) = p(fs(j)+3, j)
                jj = jj+3+1
            enddo
        endif
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        p0 = p1
        p1 = p
    enddo  ! End of time-marching loop
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! Export seismogram
    if (par%fileformat(1:2) == 'su') then
        call filename(output, par%csg_out, is, '.su')
        call write_sufile(output,csg_calc(:,1:coord%ng(is)), par%nt, coord%ng(is), par%dt, 1.0)
    !    call write_sufile_2d_offset(par%csg_out, (is-1)*coord%ngmax, is, csg_calc(:,1:coord%ng(is)), &
    !       par%nt, coord%ng(is), par%dt, 1.0, coord)
    else
        call filename(output, par%csg_out, is, '.bin')
        call write_binfile(output,csg_calc(:,1:coord%ng(is)), par%nt, coord%ng(is))
    endif
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    if (compute_residual) then
        allocate(csg_obs(par%nt,coord%ng(is)))
        allocate(dp(par%nt,coord%ngmax))    
        call read_binfile('wadi_data_window.bin', dwin, coord%ns, 2*coord%ngmax)

        if (par%fileformat(1:2) == 'su') then
            call filename(output, par%csg_in, is, '.su')
            call read_sufile(output, csg_obs, par%nt, coord%ng(is), d1, d2)
        else
            call filename(output, par%csg_in, is, '.bin')
            call read_binfile(output, csg_obs, par%nt, coord%ng(is))
        endif
        !call read_sufile(output, csg_obs, par%nt, coord%ng(is), d1, d2)
    
        if ( par%normalize == 1 ) then
            !obs_fac = max_abs(csg_obs,par%nt,coord%ng(is))
            !if (obs_fac < 1e-6) obs_fac = 1.0
            !csg_obs = csg_obs/obs_fac
    
            !calc_fac = max_abs(csg_calc,par%nt,coord%ng(is))
            !if (calc_fac < 1e-6) calc_fac = 1.0
            !csg_calc = csg_calc/calc_fac
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            do ig = 1, coord%ng(is), 1 
                it1 = int( dwin(is, 2*(ig-1)+1) )
                it2 = int( dwin(is, 2*(ig-1)+2) )            
                
                obs_fac = maxval( abs(csg_obs(it1:it2, ig)) )
                if(obs_fac < 1e-6) obs_fac = 1.d0
                csg_obs(it1:it2, ig) = csg_obs(it1:it2, ig)/obs_fac
                
                calc_fac = maxval( abs(csg_calc(it1:it2, ig)) )
                if(calc_fac < 1e-6) calc_fac = 1.d0
                csg_calc(it1:it2, ig) = csg_calc(it1:it2, ig)/calc_fac
            enddo 
        endif
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        dp = 0.0
        do ig = 1, coord%ng(is), 1
            if(abs(is-ig)>=8) then ! muting near offset traces for real data
                it1 = int( dwin(is, 2*(ig-1)+1) )
                it2 = int( dwin(is, 2*(ig-1)+2) )            
                dp(it1:it2, ig) = csg_obs(it1:it2, ig) - csg_calc(it1:it2, ig)
            endif
        enddo 
        !temp_res = norm2(csg_obs(:,1:coord%ng(is))-csg_calc(:,1:coord%ng(is)), par%nt, coord%ng(is))
        temp_res = norm2(dp, par%nt, coord%ng(is))
        res      = res + temp_res
        write(*,*) 'shot = ', is, ';  Residual = ', temp_res
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        deallocate(csg_obs, dp)
    endif
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    if (store_boundary) then
        p_end = p1(npml+1:nz_pml-npml,npml+1:nx_pml-npml)
        p_end_minus1 = p0(npml+1:nz_pml-npml,npml+1:nx_pml-npml)
    endif
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    deallocate(p0,p1,p,csg_calc, dwin)
    999 continue

end subroutine a2d42_modeling_is

!-----------------------------------------------------------------------------------------
!forward modeling 6-th order in space, 2nd order in time

subroutine fd26_modeling_is(is, par, coord, s, c, fs, nx_pml, nz_pml, npml, damp, &
              store_boundary, boundary, p_end, p_end_minus1, compute_residual, res, temp_res, wvfield)

use io
use datatype
use math
use modeling
use source
use string

type(param),       intent(in)    :: par
type(acquisition), intent(in)    :: coord
logical,           intent(in)    :: store_boundary, compute_residual
integer,           intent(in)    :: is, fs(:), nx_pml, nz_pml, npml
real,              intent(in)    :: s(:), damp(:,:), c(:,:)
real,              intent(out)   :: boundary(:,:), p_end(:,:), p_end_minus1(:,:)
real,              intent(inout) :: res, temp_res, wvfield(:,:,:)
integer                          :: it1, it2 ! adding window to traces

! Memory allocations
allocate(p0(nz_pml,nx_pml))
allocate(p1(nz_pml,nx_pml))
allocate(p(nz_pml,nx_pml))
allocate(csg_calc(par%nt,coord%ngmax))
allocate (dwin(coord%ns, 2*coord%ngmax))

    ! Initialize arrays
    p0 = 0.0
    p1 = 0.0
    p  = 0.0
    csg_calc = 0.0
    
    ! Setup source and receiver position
    isx = npml+int(coord%xs(is)/par%dx)+1
    isz = npml+int(coord%zs(is)/par%dx)+1
    ! Source position is shifted downward 1 node if it is on the free surface.
    if (isz == fs(isx)) isz = isz + 1
    
    ! Add source to the initial pressure field
    p0(isz,isx) = s(1)
    p1(isz,isx) = s(2)
    
    do ig=1,coord%ng(is),1+par%skiptrace
        igx = npml + int(coord%xg(is,ig)/par%dx)+1
        igz = npml + int(coord%zg(is,ig)/par%dx)+1
        ! Geophone position is shifted downward 1 node if it is on the free surface.
        if (igz == fs(igx)) igz = igz + 1
        csg_calc(1,ig) = p0(igz,igx)
        csg_calc(2,ig) = p1(igz,igx)
    enddo
    
    wvfield(:, :, 1) = p0( npml+1:npml+par%nz, npml+1:npml+par%nx ) !wavefield saved for RTM
    wvfield(:, :, 2) = p1( npml+1:npml+par%nz, npml+1:npml+par%nx ) !wavefield saved for RTM
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    if (store_boundary) then
        jj = 0
        ! Store the 3 bottom lines
        do ix=npml+1,nx_pml-npml
            boundary(jj+1,1) = p0(nz_pml-npml  ,ix)
            boundary(jj+2,1) = p0(nz_pml-npml-1,ix)
            boundary(jj+3,1) = p0(nz_pml-npml-2,ix)
            jj = jj+2+1
        enddo
        ! Store the 3 left and 3 right lines
        do iz=npml+1,nz_pml-npml
            boundary(jj+1,1) = p0(iz,npml+1)
            boundary(jj+2,1) = p0(iz,npml+2)
            boundary(jj+3,1) = p0(iz,npml+3)
    
            boundary(jj+4,1) = p0(iz,nx_pml-npml)
            boundary(jj+5,1) = p0(iz,nx_pml-npml-1)
            boundary(jj+6,1) = p0(iz,nx_pml-npml-2)
            jj = jj+4+2
        enddo
        ! Store the 4 top lines
        do j=npml+1,nx_pml-npml
            boundary(jj+1,1) = p0(fs(j)  ,j)
            boundary(jj+2,1) = p0(fs(j)+1,j)
            boundary(jj+3,1) = p0(fs(j)+2,j)
            boundary(jj+4,1) = p0(fs(j)+3,j)
            jj = jj+3+1
        enddo
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        jj = 0
        ! Store the 3 bottom lines
        do ix=npml+1,nx_pml-npml
            boundary(jj+1,2) = p1(nz_pml-npml  ,ix)
            boundary(jj+2,2) = p1(nz_pml-npml-1,ix)
            boundary(jj+3,2) = p1(nz_pml-npml-2,ix)      
            jj = jj+2+1
        enddo
        ! Store the 3 left and 3 right lines
        do iz=npml+1,nz_pml-npml
            boundary(jj+1,2) = p1(iz,npml+1)
            boundary(jj+2,2) = p1(iz,npml+2)
            boundary(jj+3,2) = p1(iz,npml+3)
    
            boundary(jj+4,2) = p1(iz,nx_pml-npml)
            boundary(jj+5,2) = p1(iz,nx_pml-npml-1)
            boundary(jj+6,2) = p1(iz,nx_pml-npml-2)
            jj = jj+4+2
        enddo
        ! Store the 4 top lines
        do j=npml+1,nx_pml-npml
            boundary( jj+1, 2) = p1( fs(j)  , j )
            boundary( jj+2, 2) = p1( fs(j)+1, j )
            boundary( jj+3, 2) = p1( fs(j)+2, j )
            boundary( jj+4, 2) = p1( fs(j)+3, j )
            jj = jj+3+1
        enddo
    endif
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! Time marching loop
    do it=3,par%nt
    ! Update the pressure field
        !do ix=3,nx_pml-2
        !  do iz=fs(ix)+1,nz_pml-2
        do ix = 4, nx_pml-3
            do iz = fs(ix)+1, nz_pml-3
              alpha = c(iz,ix)*par%dt/par%dx
              alpha = alpha*alpha
              kappa = damp(iz,ix)*par%dt
    !          p(iz,ix) = (2.0+2.0*c1_2nd_order*alpha-kappa)*p1(iz,ix) - &
    !                     (1.0-kappa)*p0(iz,ix) + alpha* &
    !                     (c2_2nd_order*(p1(iz,ix+1)+p1(iz,ix-1)+p1(iz+1,ix)+p1(iz-1,ix)) + &
    !                      c3_2nd_order*(p1(iz,ix+2)+p1(iz,ix-2)+p1(iz+2,ix)+p1(iz-2,ix)))
              p(iz,ix) = (2.0-kappa)*p1(iz,ix) - (1.0-kappa)*p0(iz,ix) + &
                           alpha*prefact*( 2.0*c1_6th*p1(iz,ix) + &
                           c2_6th*( p1(iz,ix+1)+p1(iz,ix-1)+p1(iz+1,ix)+p1(iz-1,ix) ) + &
                           c3_6th*( p1(iz,ix+2)+p1(iz,ix-2)+p1(iz+2,ix)+p1(iz-2,ix) ) + &
                           c4_6th*( p1(iz,ix+3)+p1(iz,ix-3)+p1(iz+3,ix)+p1(iz-3,ix) ) )
            enddo
        enddo

        wvfield(:,:,it) = (p( npml+1:npml+par%nz, npml+1:npml+par%nx ) - & !wavefield saved for RTM
                           p1( npml+1:npml+par%nz, npml+1:npml+par%nx) )/par%dt
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        ! Add source
        beta = c(isz,isx)*par%dt
        beta = beta*beta
        p(isz,isx) = p(isz,isx) + beta*s(it)
    
        ! Free surface
        do ix=1,nx_pml
            p(fs(ix)-1,ix) = -p(fs(ix)+1,ix)
            p(fs(ix)-2,ix) = -p(fs(ix)+2,ix); !depends on the FD scheme
        enddo
    
        ! Output pressure seismogram
        do ig=1,coord%ng(is),1+par%skiptrace
            igx = npml + int(coord%xg(is,ig)/par%dx)+1
            igz = npml + int(coord%zg(is,ig)/par%dx)+1
            if (igz == fs(igx)) igz = igz + 1
            csg_calc(it,ig) = p(igz,igx)
        enddo
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        if (store_boundary) then
            jj = 0
            ! Store the 3 bottom lines
            do ix=npml+1,nx_pml-npml
                boundary(jj+1, it) = p(nz_pml-npml  , ix)
                boundary(jj+2, it) = p(nz_pml-npml-1, ix)
                boundary(jj+3, it) = p(nz_pml-npml-2, ix)
                jj = jj+2+1
            enddo
            ! Store the 3 left and 3 right lines
            do iz=npml+1,nz_pml-npml
                boundary(jj+1, it) = p(iz, npml+1)
                boundary(jj+2, it) = p(iz, npml+2)
                boundary(jj+3, it) = p(iz, npml+3)
              
                boundary(jj+4, it) = p(iz, nx_pml-npml)
                boundary(jj+5, it) = p(iz, nx_pml-npml-1)
                boundary(jj+6, it) = p(iz, nx_pml-npml-2)
                jj = jj+4+2
            enddo
            ! Store the 4 top lines
            do j=npml+1,nx_pml-npml
                boundary(jj+1, it) = p(fs(j)  , j)
                boundary(jj+2, it) = p(fs(j)+1, j)
                boundary(jj+3, it) = p(fs(j)+2, j)
                boundary(jj+4, it) = p(fs(j)+3, j)
                jj = jj+3+1
            enddo
        endif
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        p0 = p1
        p1 = p
    enddo  ! End of time-marching loop
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! Export seismogram
    if (par%fileformat(1:2) == 'su') then
        call filename(output, par%csg_out, is, '.su')
        call write_sufile(output,csg_calc(:,1:coord%ng(is)), par%nt, coord%ng(is), par%dt, 1.0)
    !    call write_sufile_2d_offset( par%csg_out, (is-1)*coord%ngmax, is, csg_calc(:,1:coord%ng(is)), &
    !                                 par%nt, coord%ng(is), par%dt, 1.0, coord )
    else
        call filename(output, par%csg_out, is, '.bin')
        call write_binfile(output,csg_calc(:,1:coord%ng(is)), par%nt, coord%ng(is))
    endif
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    if (compute_residual) then
        allocate(csg_obs(par%nt,coord%ng(is)))
        allocate(dp(par%nt,coord%ngmax))    
        call read_binfile('wadi_data_window.bin', dwin, coord%ns, 2*coord%ngmax)

        if (par%fileformat(1:2) == 'su') then
            call filename(output, par%csg_in, is, '.su')
            call read_sufile(output, csg_obs, par%nt, coord%ng(is), d1, d2)
        else
            call filename(output, par%csg_in, is, '.bin')
            call read_binfile(output, csg_obs, par%nt, coord%ng(is))
        endif
    
        if ( par%normalize == 1 ) then
            !obs_fac = max_abs(csg_obs,par%nt,coord%ng(is))
            !if (obs_fac < 1e-6) obs_fac = 1.0
            !csg_obs = csg_obs/obs_fac
    
            !calc_fac = max_abs(csg_calc,par%nt,coord%ng(is))
            !if (calc_fac < 1e-6) calc_fac = 1.0
            !csg_calc = csg_calc/calc_fac
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            do ig = 1, coord%ng(is), 1 
                it1 = int( dwin(is, 2*(ig-1)+1) )
                it2 = int( dwin(is, 2*(ig-1)+2) )            
                
                obs_fac = maxval( abs(csg_obs(it1:it2, ig)) )
                if(obs_fac < 1e-6) obs_fac = 1.d0
                csg_obs(it1:it2, ig) = csg_obs(it1:it2, ig)/obs_fac
                
                calc_fac = maxval( abs(csg_calc(it1:it2, ig)) )
                if(calc_fac < 1e-6) calc_fac = 1.d0
                csg_calc(it1:it2, ig) = csg_calc(it1:it2, ig)/calc_fac
            enddo 
        endif
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        dp = 0.0
        do ig = 1, coord%ng(is), 1 
            if(abs(is - ig) >=8) then   !muting near offset traces for land data
                it1 = int( dwin(is, 2*(ig-1)+1) ) 
                !it1 = 1    ! first arrival matching? 
                it2 = int( dwin(is, 2*(ig-1)+2) )            
                dp(it1:it2, ig) = csg_obs(it1:it2, ig) - csg_calc(it1:it2, ig)
            endif
        enddo 
        !temp_res = norm2(csg_obs(:,1:coord%ng(is))-csg_calc(:,1:coord%ng(is)), par%nt, coord%ng(is))
        temp_res = norm2( dp, par%nt, coord%ng(is) )
        res      = res + temp_res
        write(*,*) 'shot = ', is, ';  Residual = ', temp_res
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        deallocate(csg_obs, dp)
    endif
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    if (store_boundary) then
        p_end = p1(npml+1:nz_pml-npml,npml+1:nx_pml-npml)
        p_end_minus1 = p0(npml+1:nz_pml-npml,npml+1:nx_pml-npml)
    endif
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    deallocate( p0, p1, p, csg_calc, dwin )
    999 continue

end subroutine fd26_modeling_is

!---------------------------------------------------------------------------------!

subroutine a2d42_recon_it(is, it, par, coord, s, c, fs, boundary, p0, p1, p)

use io
use datatype
use modeling
use string
use pml

type(param), intent(in)       :: par
type(acquisition), intent(in) :: coord
integer, intent(in)           :: fs(:), is, it
real, intent(in)              :: s(:), c(:,:)
real, intent(inout)           :: p0(:,:), p1(:,:), p(:,:)
real, intent(in)              :: boundary(:,:)

    ! Setup source and receiver position
    isx = int(coord%xs(is)/par%dx)+1
    isz = int(coord%zs(is)/par%dx)+1
    if (isz == fs(isx)) isz = isz + 1
    
    ! Subtract source
    beta = c(isz,isx)*par%dt
    beta = beta*beta
    p0(isz,isx) = p0(isz,isx) - beta*s(it+2)
    
    ! Update the pressure field
    do ix=4,par%nx-3
        do iz=fs(ix+npml)-npml+4,par%nz-3
    !do ix=3,par%nx-2  
    !    do iz=fs(ix+npml)-npml+3, par%nz-2
            alpha = c(iz,ix)*par%dt/par%dx
            alpha = alpha*alpha
      !      p(iz,ix) = (2.0+2.0*c1_2nd_order*alpha)*p1(iz,ix) - p0(iz,ix) + alpha* &
      !                 (c2_2nd_order*(p1(iz,ix+1)+p1(iz,ix-1)+p1(iz+1,ix)+p1(iz-1,ix)) + &
      !                 c3_2nd_order*(p1(iz,ix+2)+p1(iz,ix-2)+p1(iz+2,ix)+p1(iz-2,ix)))
            p(iz,ix) = (2.0)*p1(iz,ix) - p0(iz,ix) + & 
                       alpha*prefact*( 2.0*c1_6th*p1(iz,ix) + & 
                       c2_6th*( p1(iz,ix+1)+p1(iz,ix-1)+p1(iz+1,ix)+p1(iz-1,ix) ) + & 
                       c3_6th*( p1(iz,ix+2)+p1(iz,ix-2)+p1(iz+2,ix)+p1(iz-2,ix) ) + & 
                       c4_6th*( p1(iz,ix+3)+p1(iz,ix-3)+p1(iz+3,ix)+p1(iz-3,ix) ) ) 
        enddo
    enddo
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
    jj = 0
    !bottom bc
    do ix=1,par%nx
        p(par%nz  ,ix) = boundary(jj+1,it)
        p(par%nz-1,ix) = boundary(jj+2,it)
        p(par%nz-2,ix) = boundary(jj+3,it)
        jj = jj+2+1
    enddo
    
    !left and right bc
    do iz=1,par%nz
        p(iz,1) = boundary(jj+1,it)
        p(iz,2) = boundary(jj+2,it)
        p(iz,3) = boundary(jj+3,it)
    
        p(iz,par%nx) = boundary(jj+4,it)
        p(iz,par%nx-1) = boundary(jj+5,it)
        p(iz,par%nx-2) = boundary(jj+6,it)  
        jj = jj+4+2
    enddo
    
    !top bc
    do j=1,par%nx
        p(fs(j+npml)-npml  ,j) = boundary(jj+1,it)
        p(fs(j+npml)-npml+1,j) = boundary(jj+2,it)
        p(fs(j+npml)-npml+2,j) = boundary(jj+3,it)
        p(fs(j+npml)-npml+3,j) = boundary(jj+4,it)
        jj = jj+3+1
    enddo
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
    
    p0 = p1
    p1 = p

end subroutine a2d42_recon_it

!-----------------------------------------------------------------------------------------

subroutine a2d42_back_it(is, it, par, coord, s, c, fs, nx_pml, nz_pml, npml, damp, p0, p1, p)

use io
use datatype
use math
use modeling

type(param), intent(in)       :: par
type(acquisition), intent(in) :: coord
integer, intent(in)           :: fs(:), nx_pml, nz_pml, npml, is, it
real, intent(in)              :: damp(:,:), c(:,:), s(:,:)
real, intent(inout)           :: p0(:,:), p1(:,:), p(:,:)

    ! Setup source and receiver position
    isx = npml+int(coord%xs(is)/par%dx)+1
    isz = npml+int(coord%zs(is)/par%dx)+1
    if (isz == fs(isx)) isz = isz + 1
    
    ! Update the pressure field
    do ix=4,nx_pml-3
      do iz=fs(ix)+1,nz_pml-3
        alpha = c(iz,ix)*par%dt/par%dx
        alpha = alpha*alpha
        kappa = damp(iz,ix)*par%dt
    !    p(iz,ix) = (2.0+2.0*c1_2nd_order*alpha-kappa)*p1(iz,ix) - &
    !               (1.0-kappa)*p0(iz,ix) + alpha* &
    !               (c2_2nd_order*(p1(iz,ix+1)+p1(iz,ix-1)+p1(iz+1,ix)+p1(iz-1,ix)) + &
    !                c3_2nd_order*(p1(iz,ix+2)+p1(iz,ix-2)+p1(iz+2,ix)+p1(iz-2,ix)))
         p(iz,ix) = (2.0-kappa)*p1(iz,ix) - (1.0-kappa)*p0(iz,ix) + & 
                         alpha*prefact*( 2.0*c1_6th*p1(iz,ix) + & 
                         c2_6th*( p1(iz,ix+1)+p1(iz,ix-1)+p1(iz+1,ix)+p1(iz-1,ix) ) + & 
                         c3_6th*( p1(iz,ix+2)+p1(iz,ix-2)+p1(iz+2,ix)+p1(iz-2,ix) ) + & 
                         c4_6th*( p1(iz,ix+3)+p1(iz,ix-3)+p1(iz+3,ix)+p1(iz-3,ix) ) ) 
      enddo
    enddo
    
    ! Add source
    do ig=1,coord%ng(is),1+par%skiptrace
        igx = npml + int(coord%xg(is,ig)/par%dx)+1
        igz = npml + int(coord%zg(is,ig)/par%dx)+1
        if (igz == fs(igx)) igz = igz + 1
    
        beta = c(igz,igx)*par%dt
        beta = beta*beta
        p(igz,igx) = p(igz,igx) + beta*s(it,ig)
    enddo
    
    ! Free surface
    do ix=1,nx_pml
        p(fs(ix)-1,ix) = -p(fs(ix)+1, ix)
        p(fs(ix)-2,ix) = -p(fs(ix)+2, ix)
    enddo
    
    p0 = p1
    p1 = p

end subroutine a2d42_back_it

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine fd26_back_it(is, it, par, coord, s, c, fs, nx_pml, nz_pml, npml, damp, p0, p1, p)

use io
use datatype
use math
use modeling

type(param), intent(in)       :: par
type(acquisition), intent(in) :: coord
integer, intent(in)           :: fs(:), nx_pml, nz_pml, npml, is, it
real, intent(in)              :: damp(:,:), c(:,:), s(:,:)
real, intent(inout)           :: p0(:,:), p1(:,:), p(:,:)

    ! Setup source and receiver position
    isx = npml+int(coord%xs(is)/par%dx)+1
    isz = npml+int(coord%zs(is)/par%dx)+1
    if (isz == fs(isx)) isz = isz + 1
    
    ! Update the pressure field
    do ix=4,nx_pml-3
      do iz=fs(ix)+1,nz_pml-3
        alpha = c(iz,ix)*par%dt/par%dx
        alpha = alpha*alpha
        kappa = damp(iz,ix)*par%dt
    !    p(iz,ix) = (2.0+2.0*c1_2nd_order*alpha-kappa)*p1(iz,ix) - &
    !               (1.0-kappa)*p0(iz,ix) + alpha* &
    !               (c2_2nd_order*(p1(iz,ix+1)+p1(iz,ix-1)+p1(iz+1,ix)+p1(iz-1,ix)) + &
    !                c3_2nd_order*(p1(iz,ix+2)+p1(iz,ix-2)+p1(iz+2,ix)+p1(iz-2,ix)))
         p(iz,ix) = (2.0-kappa)*p1(iz,ix) - (1.0-kappa)*p0(iz,ix) + & 
                         alpha*prefact*( 2.0*c1_6th*p1(iz,ix) + & 
                         c2_6th*( p1(iz,ix+1)+p1(iz,ix-1)+p1(iz+1,ix)+p1(iz-1,ix) ) + & 
                         c3_6th*( p1(iz,ix+2)+p1(iz,ix-2)+p1(iz+2,ix)+p1(iz-2,ix) ) + & 
                         c4_6th*( p1(iz,ix+3)+p1(iz,ix-3)+p1(iz+3,ix)+p1(iz-3,ix) ) ) 
      enddo
    enddo
    
    ! Add source
    do ig=1,coord%ng(is),1+par%skiptrace
        igx = npml + int(coord%xg(is,ig)/par%dx)+1
        igz = npml + int(coord%zg(is,ig)/par%dx)+1
        if (igz == fs(igx)) igz = igz + 1
    
        beta = c(igz,igx)*par%dt
        beta = beta*beta
        p(igz,igx) = p(igz,igx) + beta*(s(it,ig) -s(it-1,ig) )/par%dt
    enddo
    
    ! Free surface
    do ix=1,nx_pml
        p(fs(ix)-1,ix) = -p(fs(ix)+1, ix)
        p(fs(ix)-2,ix) = -p(fs(ix)+2, ix)
    enddo
    
    p0 = p1
    p1 = p

end subroutine fd26_back_it

!-----------------------------------------------------------------------------------------

subroutine staggered42_modeling_is(is, par, coord, s, c, den, fs, nx_pml, nz_pml, npml, damp, &
                              store_boundary, boundary, p_end, u_end, w_end, compute_residual, res)
use io
use datatype
use math
use modeling
use source
use string

type(param), intent(in)       :: par
type(acquisition), intent(in) :: coord
logical, intent(in)           :: store_boundary, compute_residual
integer, intent(in)           :: is, fs(:), nx_pml, nz_pml, npml
real, intent(in)              :: damp(:,:), c(:,:), den(:,:)
real, intent(inout)           :: s(:)
real, intent(out)             :: boundary(:,:), p_end(:,:), u_end(:,:), w_end(:,:)
real, intent(inout)           :: res

! Memory allocations
allocate(u(nz_pml,nx_pml-1))
allocate(w(nz_pml-1,nx_pml))
allocate(p(nz_pml,nx_pml))
allocate(csg_calc(par%nt,coord%ngmax))

u = 0.0
w = 0.0
p = 0.0
csg_calc = 0.0

! Setup source and receiver position
isx = npml+int(coord%xs(is)/par%dx)+1
isz = npml+int(coord%zs(is)/par%dx)+1
if (isz == fs(isx)) isz = isz + 1

! Time marching loop
do it=2,par%nt
  ! Update the pressure field
  do ix=3,nx_pml-2
    do iz=fs(ix)+1,nz_pml-2
      alpha = den(iz,ix)*c(iz,ix)*c(iz,ix)*par%dt/(par%dx)
      kappa = damp(iz,ix)*par%dt
      p(iz,ix) = (1.0-kappa)*p(iz,ix) - alpha* &
                (c1_staggered*(u(iz,ix)-u(iz,ix-1) + w(iz,ix)-w(iz-1,ix)) + &
                 c2_staggered*(u(iz,ix+1)-u(iz,ix-2) + w(iz+1,ix)-w(iz-2,ix)))
    enddo
  enddo

  ! Add source
  beta = par%dt
  p(isz,isx) = p(isz,isx) + beta*s(it)

  ! Free surface
  if (par%free_surface == 1) then
    p(fs-1,:) = -p(fs+1,:)
  endif

  ! Update horizontal particle velocity: u
  do ix=2,nx_pml-2
    do iz=fs(ix),nz_pml-1
      alpha = par%dt/(den(iz,ix)*par%dx)
      kappa = 0.5*(damp(iz,ix)+damp(iz,ix+1))*par%dt
      u(iz,ix) = (1.0-kappa)*u(iz,ix) - alpha*(c1_staggered*(p(iz,ix+1)-p(iz,ix)) &
                                              +c2_staggered*(p(iz,ix+2)-p(iz,ix-1)))
    enddo
  enddo
  ! Update verticle particle velocity: w
  do ix=2,nx_pml-1
    do iz=fs(ix),nz_pml-2
      alpha = par%dt/(den(iz,ix)*par%dx)
      kappa = 0.5*(damp(iz,ix)+damp(iz+1,ix))*par%dt
      w(iz,ix) = (1.0-kappa)*w(iz,ix) - alpha*(c1_staggered*(p(iz+1,ix)-p(iz,ix)) &
                                              +c2_staggered*(p(iz+2,ix)-p(iz-1,ix)))
    enddo
  enddo

  ! Free surface boundary condition for velocity field
  if (par%free_surface.eq.1) then
    do ix=1,nx_pml
      w(fs(ix)-1,ix) = w(fs(ix),ix)
    enddo
  endif

  ! Output pressure seismogram
  do ig=1,coord%ng(is)
    igx = npml + int(coord%xg(is,ig)/par%dx)+1
    igz = npml + int(coord%zg(is,ig)/par%dx)+1
    if (igz == fs(igx)) igz = igz + 1
    if (par%data(1:8) == 'pressure') then
      csg_calc(it,ig) = p(igz,igx)
    else
      csg_calc(it,ig) = w(igz,igx)
    endif
  enddo

  if (store_boundary) then
    jj = 0
    ! Bottom wavefields
    do j=npml+1,nx_pml-npml
      boundary(jj+1,it) = p(nz_pml-npml  ,j)
      boundary(jj+2,it) = p(nz_pml-npml-1,j)
      boundary(jj+3,it) = w(nz_pml-npml-1,j)
      jj = jj+3
    enddo
    ! Left and right wavefields
    do j=npml+1,nz_pml-npml
      boundary(jj+1,it) = p(j,npml+1)
      boundary(jj+2,it) = p(j,npml+2)
      boundary(jj+3,it) = u(j,npml+1)
      boundary(jj+4,it) = p(j,nx_pml-npml)
      boundary(jj+5,it) = p(j,nx_pml-npml-1)
      boundary(jj+6,it) = u(j,nx_pml-npml-1)
      jj = jj+6
    enddo
    ! Top wavefields
    do j=npml+1,nx_pml-npml
      boundary(jj+1,it) = p(fs(j)  ,j)
      boundary(jj+2,it) = p(fs(j)+1,j)
      boundary(jj+3,it) = p(fs(j)+2,j)
      boundary(jj+4,it) = u(fs(j)  ,j)
      boundary(jj+5,it) = u(fs(j)+1,j)
      boundary(jj+6,it) = w(fs(j)  ,j)
      jj = jj+6
    enddo
  endif
enddo  ! End of time-marching loop

! Export seismogram
if (par%fileformat(1:2) == 'su') then
  call filename(output, par%csg_out, is, '.su')
  call write_sufile(output, csg_calc(:,1:coord%ng(is)), par%nt, coord%ng(is), par%dt_out, 1.0, &
    coord%xs(is), coord%xg(is,:))
else
  call filename(output, par%csg_out, is, '.bin')
  call write_binfile(output, csg_calc(:,1:coord%ng(is)), par%nt, coord%ng(is))
endif

if (store_boundary) then
  p_end = p(npml+1:nz_pml-npml,npml+1:nx_pml-npml)
  u_end = u(npml+1:nz_pml-npml,npml+1:nx_pml-npml)
  w_end = w(npml+1:nz_pml-npml,npml+1:nx_pml-npml)
endif

if (compute_residual) then
  allocate(csg_obs(par%nt,coord%ng(is)))

  if (par%fileformat(1:2) == 'su') then
    call filename(output, par%csg_in, is, '.su')
  else
    call filename(output, par%csg_in, is, '.bin')
  endif
  call read_sufile(output, csg_obs, par%nt, coord%ng(is), d1, d2)

  ! Normalize the observed and calculated wavefields
  if (par%normalize == 1) then
    obs_fac = max_abs(csg_obs, par%nt, coord%ng(is))
    if (abs(obs_fac) < 1.0E-10) obs_fac = 1.0
    csg_obs = csg_obs/obs_fac

    calc_fac = max_abs(csg_calc, par%nt, coord%ng(is))
    if (abs(calc_fac) < 1.0E-10) calc_fac = 1.0
    csg_calc = csg_calc/calc_fac
  endif

  res = res + norm2(csg_obs(:,1:coord%ng(is))-csg_calc(:,1:coord%ng(is)), par%nt, coord%ng(is))
  deallocate(csg_obs)
endif

deallocate(u,w,p,csg_calc)
999 continue

end subroutine staggered42_modeling_is

!-----------------------------------------------------------------------------------------

subroutine staggered42_recon_it(is, it, par, coord, s, c, den, fs, boundary, p, u, w)

use io
use datatype
use modeling
use pml
use string

type(param), intent(in)       :: par
type(acquisition), intent(in) :: coord
integer, intent(in)           :: fs(:), is, it
real, intent(in)              :: s(:), c(:,:), den(:,:)
real, intent(inout)           :: p(:,:), u(:,:), w(:,:)
real, intent(in)              :: boundary(:,:)

! Setup source and receiver position
isx = int(coord%xs(is)/par%dx)+1
isz = int(coord%zs(is)/par%dx)+1

! Update horizontal particle velocity: u
do ix=2,par%nx-2
  do iz=fs(ix+npml)-npml+1,par%nz-1
    alpha = par%dt/(den(iz,ix)*par%dx)
    u(iz,ix) = u(iz,ix) + alpha*(c1_staggered*(p(iz,ix+1)-p(iz,ix)) &
                                +c2_staggered*(p(iz,ix+2)-p(iz,ix-1)))
  enddo
enddo
! Update verticle particle velocity: w
do ix=2,par%nx-1
  do iz=fs(ix+npml)-npml+1,par%nz-2
    alpha = par%dt/(den(iz,ix)*par%dx)
    w(iz,ix) = w(iz,ix) + alpha*(c1_staggered*(p(iz+1,ix)-p(iz,ix)) &
                                +c2_staggered*(p(iz+2,ix)-p(iz-1,ix)))
  enddo
enddo

! Insert boundary values
jj = 0
do j=1,par%nx
  w(par%nz-1,j) = boundary(jj+3,it-1)
  jj = jj+3
enddo
do j=1,par%nz
  u(j,1       ) = boundary(jj+3,it-1)
  u(j,par%nx-1) = boundary(jj+6,it-1)
  jj = jj+6
enddo

! Subtract source
if (isx > 2 .and. isx < par%nx-1 .and. isz > 3 .and. isz < par%nz-1) then
  beta = par%dt
  if (isz .gt. fs(isx+npml)-npml) then
    p(isz,isx) = p(isz,isx) - beta*s(it+1)
  else
    p(fs(isx+npml)-npml+1,isx) = p(fs(isx+npml)-npml+1,isx) - beta*s(it+1)
  endif
endif

! Update the pressure field
do ix=3,par%nx-2
  do iz=fs(ix+npml)-npml+2,par%nz-2
    alpha = den(iz,ix)*c(iz,ix)*c(iz,ix)*par%dt/(par%dx)
    p(iz,ix) = p(iz,ix) + alpha* &
              (c1_staggered*(u(iz,ix)-u(iz,ix-1) + w(iz,ix)-w(iz-1,ix)) + &
               c2_staggered*(u(iz,ix+1)-u(iz,ix-2) + w(iz+1,ix)-w(iz-2,ix)))
  enddo
enddo

jj=0
do j=1,par%nx
  p(par%nz  ,j) = boundary(jj+1,it-1)
  p(par%nz-1,j) = boundary(jj+2,it-1)
  jj=jj+3
enddo
do j=1,par%nz
  p(j,1)        = boundary(jj+1,it-1)
  p(j,2)        = boundary(jj+2,it-1)
  p(j,par%nx  ) = boundary(jj+4,it-1)
  p(j,par%nx-1) = boundary(jj+5,it-1)
  jj=jj+6
enddo
do j=1,par%nx
  p(fs(j+npml)-npml  ,j) = boundary(jj+1,it-1)
  p(fs(j+npml)-npml+1,j) = boundary(jj+2,it-1)
  p(fs(j+npml)-npml+2,j) = boundary(jj+3,it-1)
  jj=jj+6
enddo

end subroutine staggered42_recon_it

!-----------------------------------------------------------------------------------------

subroutine staggered42_back_it(is, it, par, coord, s, c, den, fs, nx_pml, nz_pml, npml, damp, p, u, w)

use io
use datatype
use math
use modeling

type(param), intent(in)       :: par
type(acquisition), intent(in) :: coord
integer, intent(in)           :: fs(:), nx_pml, nz_pml, npml, is, it
real, intent(in)              :: damp(:,:), c(:,:), s(:,:), den(:,:)
real, intent(inout)           :: p(:,:), u(:,:), w(:,:)

! Setup source and receiver position
isx = npml+int(coord%xs(is)/par%dx)+1
isz = npml+int(coord%zs(is)/par%dx)+1

! Update horizontal particle velocity: u
do ix=2,nx_pml-2
  do iz=2,nz_pml-1
    alpha = par%dt/(den(iz,ix)*par%dx)
    kappa = 0.5*(damp(iz,ix)+damp(iz,ix+1))*par%dt
    u(iz,ix) = (1.0-kappa)*u(iz,ix) - alpha*(c1_staggered*(p(iz,ix+1)-p(iz,ix)) &
                                              +c2_staggered*(p(iz,ix+2)-p(iz,ix-1)))
  enddo
enddo
! Update verticle particle velocity: w
do ix=2,nx_pml-1
  do iz=2,nz_pml-2
    alpha = par%dt/(den(iz,ix)*par%dx)
    kappa = 0.5*(damp(iz,ix)+damp(iz+1,ix))*par%dt
    w(iz,ix) = (1.0-kappa)*w(iz,ix) - alpha*(c1_staggered*(p(iz+1,ix)-p(iz,ix)) &
                                              +c2_staggered*(p(iz+2,ix)-p(iz-1,ix)))
  enddo
enddo

! Free surface boundary condition for velocity field
if (par%free_surface.eq.1) then
  do ix=2,nx_pml-1
    w(fs(ix)-1,ix) = w(fs(ix),ix)
  enddo
endif

! Update the pressure field
do ix=3,nx_pml-2
  do iz=3,nz_pml-2
    alpha = den(iz,ix)*c(iz,ix)*c(iz,ix)*par%dt/(par%dx)
    kappa = damp(iz,ix)*par%dt
    p(iz,ix) = (1.0-kappa)*p(iz,ix) - alpha* &
                (c1_staggered*(u(iz,ix)-u(iz,ix-1) + w(iz,ix)-w(iz-1,ix)) + &
                 c2_staggered*(u(iz,ix+1)-u(iz,ix-2) + w(iz+1,ix)-w(iz-2,ix)))
  enddo
enddo

! Add source
do ig=1,coord%ng(is)
  igx = npml + int(coord%xg(is,ig)/par%dx)+1
  igz = npml + int(coord%zg(is,ig)/par%dx)+1
  beta = par%dt

  ! Geophone is shifted downward 1 node if it is on the free surface.
  if (par%data(1:8) == 'pressure') then
    if (igz .gt. fs(igx)) then
      p(igz,igx) = p(igz,igx) + beta*s(it,ig)
    else
      p(fs(igx)+1,igx) = p(fs(igx)+1,igx) + beta*s(it,ig)
    endif
  else
    if (igz .gt. fs(igx)) then
      w(igz,igx) = w(igz,igx) + beta*s(it,ig)
    else
      w(fs(igx)+1,igx) = w(fs(igx)+1,igx) + beta*s(it,ig)
    endif
  endif
enddo

! Free surface
do ix=1,nx_pml
!  p(fs(ix),ix) = 0.0
  p(fs(ix)-1,ix) = -p(fs(ix)+1,ix)
enddo

end subroutine staggered42_back_it

!-----------------------------------------------------------------------------------------

subroutine update_slowness(par,npml,c,dc,c_new)

use datatype
use global
use io
use math

type(param), intent(in) :: par
real, intent(in)        :: c(:,:), dc(:,:)
real, intent(out)       :: c_new(:,:)
integer, intent(in)     :: npml
integer                 :: ix1, ix2, iz1, iz2, nx_pml, nz_pml, i, j
real                    :: c_old_min, c_old_max

    nx_pml = par%nx + 2*npml
    nz_pml = par%nz + 2*npml
    ix1 = npml+1
    ix2 = nx_pml-npml
    iz1 = npml+1
    iz2 = nz_pml-npml
    
    c_new(iz1:iz2, ix1:ix2) = c(iz1:iz2, ix1:ix2) + dc
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    if (par%vconstraint > 0.0) then
        do i=1,par%nx
            do j=1,par%nz
                c_old_min = c(npml+j,npml+i)*(1.0-par%vconstraint/100.0)
                c_old_max = c(npml+j,npml+i)*(1.0+par%vconstraint/100.0)
                if (c_new(npml+j,npml+i) > c_old_max) c_new(npml+j,npml+i) = c_old_max
                if (c_new(npml+j,npml+i) < c_old_min) c_new(npml+j,npml+i) = c_old_min
            enddo
        enddo
    endif
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    do i=1,par%nx
        do j=1,par%nz
            if (c_new(npml+j,npml+i) > 1.0/par%vmin) c_new(npml+j,npml+i) = 1.0/par%vmin
            if (c_new(npml+j,npml+i) < 1.0/par%vmax) c_new(npml+j,npml+i) = 1.0/par%vmax
        enddo
    enddo
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! Extrapolate velocity in PML regions
    do i=1,npml
        c_new(iz1:iz2,i) = c_new(iz1:iz2,npml+1)
        c_new(iz1:iz2,ix2+i) = c_new(iz1:iz2,nx_pml-npml)
    enddo
    do i=1,npml
        c_new(i,:) = c_new(npml+1,:)
        c_new(iz2+i,:) = c_new(nz_pml-npml,:)
    enddo
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    if (par%smoothvel > 0) then
        call smooth(c_new, nz_pml, nx_pml, par%smoothvel)
    endif

end subroutine update_slowness

!-----------------------------------------------------------------------------------------

end module a2d

