! module for 2D applications including forward modeling, RTM and FWI
!
module apps2d

use a2d
use datatype

implicit none

type(param),       private              :: par
type(acquisition), private              :: coord
logical,           private              :: store_boundary, message
integer,           private              :: ix, iz, it, isx, isz, igx, igz, is, ig, it1,it2, &
                                           is1, is2, iter, nsearch, nsearch_max
integer,           private, allocatable :: fs(:), izwb(:)
real,              private              :: res, res_new, x2, y2, x3, y3, sx, sz, gx, gz, &
                                           alpha, beta, c1, c2, c3, kappa, d1, d2, dsx, dgx, &
                                           obs_fac, calc_fac, step_min, res_min
real,              private, allocatable :: s(:), c(:,:), den(:,:), csg_obs(:,:), &
                                           p_end(:,:), u_end(:,:), w_end(:,:), &
                                           p_end_minus1(:,:), p_recon(:,:), p1_bk(:,:), &
                                           image(:,:), image_sum(:,:), p0_bk(:,:), &
                                           p_bk(:,:), u_bk(:,:), w_bk(:,:), boundary(:,:), &
                                           slow(:,:), slow_trial(:,:), g(:,:),g_temp(:,:), ds(:,:), &
                                           dp(:,:), d(:,:), g_old(:,:), csg_calc(:,:), dd(:,:)

contains

!-----------------------------------------------------------------------------------------
subroutine forward_modeling(parfile, method)

use io
use math
use modeling
use mmi_mpi
use pml
use source

character(len=*), intent(in) :: parfile, method

call start_mpi

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!    if(rank == 0) then
!        write (*,*) 'This is rank = ', rank
!    elseif(rank == 1) then
!        write(*,*) 'I am rank = ', rank
!    else
        
!    endif
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

! Read input parameters
call readparamfile(parfile, par)
! PML setting
call init_pml(par%nx, par%nz, par%npml)
! Read acquisition geometry data
call readcoordfile(par%coordfile, coord)

! Memory allocations
allocate(s(par%nt))
allocate(fs(nx_pml))
allocate(c(nz_pml,nx_pml), den(nz_pml,nx_pml))
!allocate(boundary(par%nx*5+par%nz*4,par%nt)) !boundary related to Laplacian operator
allocate(boundary(par%nx*7+par%nz*6,par%nt)) !boundary related to Laplacian operator
allocate(p_end(par%nz,par%nx))
allocate(p_end_minus1(par%nz,par%nx))

! Set up free surface
call free_surface(par, fs, npml)
! Read velocity model
call readvelfile(par,c,npml,nx_pml,nz_pml)
! Density setup
if (par%variable_density == 1) then
  if (par%densityfile(1:3) == 'n/a') then
    call compute_density(par%variable_density, c, den, nx_pml, nz_pml)
  else
    call read_densityfile(par,c,den,npml,nx_pml,nz_pml)
  endif
else
  den = 1.0
endif

! Determine minimum velocity
par%cmin = min_value(c(iz1:iz2,ix1:ix2),par%nz,par%nx,fs-npml)
if (rank == 0) write(*,*) 'Minimum velocity = ', par%cmin
! Setup PML damping coefficient
call setup_pml(par%dx, par%cmin)
! Setup Ricker source wavelet
call getsource(par, s)

if (rank.eq.0) then
  write(*,*) 'nx_pml = ', nx_pml, ', nz_pml = ', nz_pml, ', npml = ', par%npml
  write(*,*) 'f = ', par%f, ', dx = ', par%dx
  write(*,*) 'dt = ', par%dt, ', nt = ', par%nt
  write(*,*) 'xmin = ', par%xmin, ', xmax = ', par%xmax
  write(*,*) 'offset_min = ', par%offset_min, ', offset_max = ', par%offset_max
endif

! Forward modeling using dynamic load balancing
message = .true.
call modeling_mpi(method, message, par, coord, s, c, den, fs, nx_pml, nz_pml, &
                  npml, damp_global, .false., res)

deallocate(c,s,fs,boundary,damp,damp_global,p_end,p_end_minus1,den)
999 continue

call stop_mpi

end subroutine forward_modeling
!-----------------------------------------------------------------------------------------
subroutine migration_staggered(parfile)

use datatype
use io
use math
use modeling
use mmi_mpi
use pml
use source

character(len=*), intent(in) :: parfile

call start_mpi

! Read in velocity model parameters: nx, nz, dx, filename
call readparamfile(parfile,par)

! PML setting
call init_pml(par%nx, par%nz, par%npml)

! Read acquisition geometry data
call readcoordfile(par%coordfile,coord)

! Memory allocations
allocate(s(par%nt))
allocate(c(nz_pml,nx_pml),den(nz_pml,nx_pml))
allocate(fs(nx_pml))
allocate(boundary(par%nx*9+par%nz*6,par%nt))
allocate(p_end(par%nz,par%nx))
allocate(u_end(par%nz,par%nx))
allocate(w_end(par%nz,par%nx))
allocate(p_recon(par%nz,par%nx))
allocate(image(par%nz,par%nx))
allocate(image_sum(par%nz,par%nx))
allocate(p_bk(nz_pml,nx_pml))
allocate(u_bk(nz_pml,nx_pml))
allocate(w_bk(nz_pml,nx_pml))
allocate(csg_obs(par%nt,coord%ngmax))
allocate(izwb(par%nx))

! Read waterbottom file
if (par%mute_data == 1) then
  call readwaterbottomfile(par, izwb)
endif

! Set up free surface
call free_surface(par, fs, npml)

! Read velocity model
call readvelfile(par,c,npml,nx_pml,nz_pml)

! Density setup
call compute_density(par%variable_density, c, den, nx_pml, nz_pml)

! Determine minimum velocity
par%cmin = min_value(c(iz1:iz2,ix1:ix2),par%nz,par%nx)

! Setup PML damping coefficient
call setup_pml(par%dx, par%cmin)

! Setup Ricker source wavelet
call getsource(par, s)

! Initialize the migration image
image = 0.0

call get_assigned(1, coord%ns, is1, is2)

do is=is1,is2
  write(*,*) 'Process ', rank, ', shot ', is
  flush(6)
  boundary = 0.0
  call staggered42_modeling_is(is, par, coord, s, c, den, fs, nx_pml, nz_pml, npml, damp_global, &
    .true., boundary, p_end, u_end, w_end, .false., res)

  call filename(output, par%csg_in, is, '.su')
  call read_sufile(output, csg_obs, par%nt, coord%ng(is), d1, d2)

  p_bk = 0.0
  u_bk = 0.0
  w_bk = 0.0

  do it=par%nt,2,-1
    call staggered42_recon_it(is, it, par, coord, s, c(iz1:iz2,ix1:ix2), den(iz1:iz2,ix1:ix2), &
      fs, boundary, p_end, u_end, w_end)

    call staggered42_back_it(is, it, par, coord, csg_obs, c, den, fs, nx_pml, nz_pml, npml, damp_global, &
      p_bk, u_bk, w_bk)
    image = image + p_end*p_bk(iz1:iz2,ix1:ix2)
  enddo
enddo

call MPI_Barrier(MPI_COMM_WORLD, ierr)
call MPI_Allreduce(image, image_sum, par%nz*par%nx, MPI_REAL, MPI_SUM, MPI_COMM_WORLD, ierr)

! Output migration image
if (rank == 0) then
  call write_sufile('mig.su',image_sum,par%nz,par%nx,1.0,1.0)
endif

deallocate(s,c,fs,damp,damp_global,p_end,u_end,w_end,p_recon,p_bk,u_bk,w_bk,image,csg_obs)
999 continue

call stop_mpi

end subroutine migration_staggered

!-----------------------------------------------------------------------------------------
subroutine migration_2nd_order(parfile)

use datatype
use io
use math
use modeling
use mmi_mpi
use pml
use source

character(len=*), intent(in) :: parfile

!call start_mpi

! Read in velocity model parameters: nx, nz, dx, filename
call readparamfile(parfile,par)
! PML setting
call init_pml(par%nx, par%nz, par%npml)
! Read acquisition geometry data
call readcoordfile(par%coordfile,coord)
! Memory allocations
allocate(c(nz_pml,nx_pml))
allocate(fs(nx_pml))
!allocate(boundary(par%nx*5+par%nz*4,par%nt))
allocate(boundary(par%nx*7+par%nz*6,par%nt))

allocate(p_end(par%nz,par%nx))
allocate(p_end_minus1(par%nz,par%nx))
allocate(p_recon(par%nz,par%nx))
allocate(image(par%nz,par%nx))
allocate(image_sum(par%nz,par%nx))
allocate(p0_bk(nz_pml,nx_pml))
allocate(p1_bk(nz_pml,nx_pml))
allocate(p_bk(nz_pml,nx_pml))
allocate(csg_obs(par%nt,coord%ngmax))

! Set up free surface
call free_surface(par, fs, npml)
! Read velocity model
call readvelfile(par,c,npml,nx_pml,nz_pml)
! Determine minimum velocity
par%cmin = min_value(c(iz1:iz2,ix1:ix2),par%nz,par%nx)

! Setup PML damping coefficient
call setup_pml(par%dx, par%cmin)
! Setup Ricker source wavelet
allocate(s(par%nt))
call getsource(par, s)
! Initialize the migration image
image = 0.0

call get_assigned(1, coord%ns, is1, is2)

do is=is1,is2
  write(*,*) 'Process ', rank, ', shot ', is
  flush(6)
  boundary = 0.0
  call a2d42_modeling_is(is, par, coord, s, c, fs, nx_pml, nz_pml, npml, damp_global, &
    .true., boundary, p_end, p_end_minus1, .false., res)

  call filename(output, par%csg_in, is, '.su')
  call read_sufile(output, csg_obs, par%nt, coord%ng(is), d1, d2)

  p0_bk = 0.0
  p1_bk = 0.0
  p_bk  = 0.0
  do ig=1,coord%ng(is),1+par%skiptrace
    igx = npml + int(coord%xg(is,ig)/par%dx)+1
    igz = npml + int(coord%zg(is,ig)/par%dx)+1
    ! Geophone is shifted downward 1 node if it is on the free surface.
    if (igz.ne.fs(igx)) then
      p0_bk(igz,igx) = csg_obs(par%nt,ig)
      p1_bk(igz,igx) = csg_obs(par%nt-1,ig)
    else
      p0_bk(igz+1,igx) = csg_obs(par%nt,ig)
      p1_bk(igz+1,igx) = csg_obs(par%nt-1,ig)
    endif
  enddo

  do it=par%nt-2,1,-1
      call a2d42_recon_it(is, it, par, coord, s, c(npml+1:nz_pml-npml,npml+1:nx_pml-npml), &
        fs, boundary, p_end, p_end_minus1, p_recon)

      call a2d42_back_it(is, it, par, coord, csg_obs, c, fs, nx_pml, nz_pml, npml, damp_global, &
        p0_bk, p1_bk, p_bk)
      image = image + p_recon*p_bk(iz1:iz2,ix1:ix2)
  enddo
enddo
call MPI_Barrier(MPI_COMM_WORLD, ierr)
call MPI_Allreduce(image, image_sum, par%nz*par%nx, MPI_REAL, MPI_SUM, MPI_COMM_WORLD, ierr)

! Output migration image
if (rank == 0) then
  call write_sufile('mig.su',image_sum,par%nz,par%nx,par%dx/1.0e3,par%dx/1.0e3)
endif

deallocate(s,c,fs,damp,damp_global,p_end,p_end_minus1,p_recon,image)
999 continue
call stop_mpi

end subroutine migration_2nd_order

!-----------------------------------------------------------------------------------------!

subroutine slowness_inversion_2nd_order(parfile)

use datatype
use io
use math
use modeling
use mmi_mpi
use pml
use source

character(len=*), intent(in) :: parfile
logical :: ifmysteplen
character:: yn
character(len=3)   :: round
character(len=40)  :: gradfile, srcfile
!integer           :: nsng
real, allocatable  :: dwin(:,:), temp_src(:), wvfield(:,:,:)
real               :: temp_res

    message = .true.    
    call start_mpi
    
    ! Read in velocity model parameters: nx, nz, dx, filename
    call readparamfile(parfile,par)
    ! PML setting
    call init_pml(par%nx, par%nz, par%npml)
    ! Read acquisition geometry data
    call readcoordfile(par%coordfile,coord)
    ! Memory allocations
    allocate(s(par%nt))
    allocate(temp_src(par%nt))
    allocate(c(nz_pml,nx_pml))
    allocate(slow(nz_pml,nx_pml))
    allocate(slow_trial(nz_pml,nx_pml))
    allocate(ds(par%nz,par%nx))
    allocate(fs(nx_pml))
    allocate(boundary(par%nx*7+par%nz*6,par%nt))
    allocate(p_end(par%nz,par%nx))
    allocate(p_end_minus1(par%nz,par%nx))
    allocate(p_recon(par%nz,par%nx))
    allocate(p0_bk(nz_pml,nx_pml))
    allocate(p1_bk(nz_pml,nx_pml))
    allocate(p_bk(nz_pml,nx_pml))
    allocate(csg_obs(par%nt,coord%ngmax))
    allocate(csg_calc(par%nt,coord%ngmax))       
    allocate(dwin(coord%ns, 2*coord%ngmax))
    allocate( wvfield(par%nz, par%nx, par%nt) )
   
    allocate(dp(par%nt,coord%ngmax))
    allocate(g(par%nz,par%nx))
    allocate(g_temp(par%nz, par%nx))
    allocate(g_old(par%nz,par%nx))
    allocate(d(par%nz,par%nx))
    allocate(dd(par%nz,par%nx))
    
    if (par%ismarine == 1) then
        ! Read waterbottom file
        allocate(izwb(par%nx))
        if (par%waterbottomfile(1:3) == 'n/a') then
            if (par%izwb == -1) then
                izwb = 1
            else
                izwb = par%izwb
            endif
        else
            call readwaterbottomfile(par, izwb)
        endif
    endif
    
    ! Set up free surface
    call free_surface(par, fs, npml)
    ! Read velocity model
    call readvelfile(par,c,npml,nx_pml,nz_pml)
    slow = 1.0/c
    ! Determine minimum velocity
    par%cmin = min_value(c(iz1:iz2,ix1:ix2),par%nz,par%nx)
    
    ! Setup PML damping coefficient
    call setup_pml(par%dx, par%cmin)
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! Setup Ricker source wavelet
    call getsource(par, s)
    !s(1:400)      = s(80:479)     
    !s(401:par%nt) = 0.0
    !call write_binfile('src.bin', s, par%nt)
    
    s = 0.0 ! assign some natural source
    !call read_binfile('natsrc.bin', s, par%nt)
    call read_binfile('src.bin', s, par%nt)
    !s(1:400)      = s(70:469)
    !s(401:par%nt) = 0.0
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ! Log file
    if (rank == 0) then
        write(*,*) 'cmin = ', par%cmin
        flush( 6 )
        open( 1024, file=par%logfile, form='formatted' )
        open( 1025, file='res_record.txt', form='formatted' )
        !!!!!!!!!!!!!!!!!
        ! set if using a different step length
        ifmysteplen = .false.
        write(*,*) 'Do you want to use another step length searching other than back track? (y/n)'
        read(*,*) yn
        if (yn == 'y') then
           ifmysteplen = .true.
        else
           ifmysteplen = .false.
        endif
        !!!!!!!!!!!!!!!!!
    endif
    call MPI_Barrier(MPI_COMM_WORLD, ierr)
    call MPI_BCAST(ifmysteplen, 1, MPI_LOGICAL, 0, MPI_COMM_WORLD, ierr)

    call read_binfile( 'wadi_data_window.bin', dwin, coord%ns, 2*coord%ngmax)
    !write(*,*) 'max stop = ', int(max_abs(dwin, coord%ns, 2*coord%ng(1)) )  ! for testing
    call get_assigned( 1, coord%ns, is1, is2 )
    open( 7, file='res1.txt', form='formatted')
    !--------------------------------------------------------------------------------------!
    ! Waveform inversion iteration
    nsearch_max = 10
    nsearch = 0
    do iter = 1, par%itermax
        ! Initialize the migration image
        ds  = 0.0
        res = 0.0
        !~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~!   
        do is = is1,is2
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            boundary = 0.0
            call fd26_modeling_is( is, par, coord, s, c, fs, nx_pml, nz_pml, npml, damp_global, &
                                    .true., boundary, p_end, p_end_minus1, .true., res , temp_res, wvfield) 
            if( is == 1 ) then
                write(7, *)  'iter = ', iter, ', residual = ', temp_res
                flush(7)
            endif
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            if (par%fileformat(1:2) == 'su') then ! reading observed data: csg_obs
                call filename(output, par%csg_in, is, '.su')
                call read_sufile( output, csg_obs, par%nt, coord%ng(is), d1, d2 )
            else
                call filename(output, par%csg_in, is, '.bin')
                call read_binfile( output, csg_obs, par%nt, coord%ng(is) )
            endif
                                                                                          
            if (par%fileformat(1:2) == 'su') then ! reading calculated data: csg_calc
                call filename(output, par%csg_out, is, '.su')
                call read_sufile( output, csg_calc, par%nt, coord%ng(is), d1, d2 )
            else
                call filename( output, par%csg_out, is, '.bin' )
                call read_binfile( output, csg_calc, par%nt, coord%ng(is) )
            endif
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            if (par%normalize == 1) then
                ! Normalize the observed and calculated wavefields
                !obs_fac = max_abs(csg_obs, par%nt, coord%ng(is))
                !if (abs(obs_fac) < 1.0E-10) obs_fac = 1.0
                !csg_obs = csg_obs/obs_fac    
                
                !calc_fac = max_abs(csg_calc, par%nt, coord%ng(is))
                !if (abs(calc_fac) < 1.0E-10) calc_fac = 1.0
                !csg_calc = csg_calc/calc_fac
                !write(*,*) 'shot=', is, '; calc_max =', calc_fac, '; obs_max=',obs_fac
                !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                do ig = 1, coord%ng(is),1
                    it1 = int( dwin(is, 2*(ig-1) +1))
                    it2 = int( dwin(is, 2*(ig-1) +2))
       
                    obs_fac = maxval( abs(csg_obs(it1:it2, ig)) )
                    if(obs_fac < 1e-6) obs_fac = 1.d0
                    csg_obs(it1:it2, ig) = csg_obs(it1:it2, ig)/obs_fac
 
                    calc_fac = maxval( abs(csg_calc(it1:it2, ig)) )
                    if(calc_fac < 1e-6) calc_fac = 1.d0
                    csg_calc(it1:it2, ig) = csg_calc(it1:it2, ig)/calc_fac
                enddo
            endif

            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            !dp = csg_obs - csg_calc
            !!windowing early arrival
            dp = dp * 0.0
            do ig = 1,coord%ng(is),1
                if(abs(is-ig)>=8) then  ! muting near offset traces for land data
                    it1 = int(dwin(is, 2*(ig-1)+1))
                    !it1 = 1 !first arrival matches?
                    it2 = int(dwin(is, 2*(ig-1)+2))
                    dp(it1:it2,ig) = csg_obs(it1:it2,ig) - csg_calc(it1:it2,ig)
                endif
            enddo
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            p0_bk = 0.0
            p1_bk = 0.0
            p_bk  = 0.0
            do ig=1, coord%ng(is), 1+par%skiptrace
                igx = npml + int(coord%xg(is,ig)/par%dx)+1
                igz = npml + int(coord%zg(is,ig)/par%dx)+1
                if (igz == fs(igx)) igz = igz+1
                p0_bk(igz,igx) = dp(par%nt,ig)
                p1_bk(igz,igx) = dp(par%nt-1,ig)
            enddo
            do it = par%nt-2, 1, -1
                !call a2d42_recon_it( is, it, par, coord, s, c(npml+1:nz_pml-npml,npml+1:nx_pml-npml), &
                !                     fs, boundary, p_end, p_end_minus1, p_recon )                 
                call fd26_back_it( is, it, par, coord, dp, c, fs, nx_pml, nz_pml, npml, damp_global, &
                                    p0_bk, p1_bk, p_bk )
                !ds = ds + p_recon*p_bk(iz1:iz2,ix1:ix2)
                ds = ds + wvfield( :, :, it ) * p_bk( iz1:iz2, ix1:ix2 )
            enddo
        enddo
        !~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~!   
        call MPI_Barrier(MPI_COMM_WORLD, ierr)
        call MPI_Allreduce(ds, g, par%nz*par%nx, MPI_REAL, MPI_SUM, MPI_COMM_WORLD, ierr)
        call MPI_Allreduce(res, res_new, 1, MPI_REAL, MPI_SUM, MPI_COMM_WORLD, ierr)
        res_new = sqrt(res_new) 
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        ! Mute the gradient using the water bottom depth
        if ( par%ismarine == 1 ) then
            call mute2( g, par%nx, par%nz, izwb, 10 )
        endif
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        g = 2.0 * g * max_value(slow(iz1:iz2,ix1:ix2),par%nz,par%nx)/mean_abs(g,par%nz,par%nx)
            
        !dd = 2.0*g*slow(iz1:iz2,ix1:ix2); 
        !write(*,*) 'max dd=', maxval(dd),'min dd=',minval(dd)
        !write(*,*) 'max slow=', maxval(slow), 'min slow=',minval(slow)
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        if (par%smoothgrad > 0) then
            call smooth_2d_pad(g, par%nz, par%nx, par%smoothgrad)
        endif        
        if ( iter == 1  .or.  mod(iter,3) == 1 ) then
            d = g
        else
            beta = dot(g-g_old,g,par%nz,par%nx)/dot(g_old,g_old,par%nz,par%nx)
            d    = g + beta * d
        endif
        
        if ( rank == 0 ) then
            write(*,*) ' Update Gradient: g for iter = 1, d for iter > 1 '
            write(*,*) ' max gradient=', max_value(d,par%nz,par%nx), & 
                       '; min gradient=',min_value(d, par%nz,par%nx)            
            write(*, *) ' Iterataion =', iter, ', data residual =', res_new
            write(1024,*) ' Iter =', iter, ', residual =', res_new
            write(1025,*) 'Iter: ', iter, '; res: ', res_new  
            ! showing only the data residual, not for internal steps
            flush(6)
            flush(1024)
            flush(1025)
    
            ! Output waveform gradient
            write(round,'(i3)') iter
            gradfile = 'grad-'//trim(adjustl(round))//'.bin'
            !call write_sufile(gradfile, g, par%nz, par%nx, 1.0, 1.0)
            call write_binfile( gradfile, d, par%nz, par%nx ) ! d or g.
        endif
        !---------------------------------------------------------------------------!
        ! Calc a step length using back-tracking or quadratic line search
        !---------------------------------------------------------------------------!
        if ( par%line_search == 0 ) then ! line_search = 1 by default setting
            alpha    = par%step
            nsearch  = 0
            step_min = alpha
            res_min  = res_new
            !write(*,*) 'd=',d
                 
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            100  continue
            call update_slowness( par, npml, slow, alpha*d/100.0, slow_trial )  ! d is direction
            c    = 1.0/slow_trial
            call modeling_mpi( '2nd_order', .false., par, coord, s, c, den, fs, nx_pml, nz_pml, &
                                npml, damp_global, .true., y2 )
            
            if (rank == 0) then
                write(*,*)  '   step = ', alpha, ', res = ', y2
                write(1024,*) '   step = ', alpha, ', res = ', y2
                flush(6)
                flush(1024)
            endif
            
            if (y2 > 0.98*res_new) then  ! 0.95 means expecting 5% convergence rate
                if (res_min > y2) then
                    res_min  = y2
                    step_min = alpha
                endif
                alpha   = alpha/2.0
                nsearch = nsearch+1

                if( nsearch>nsearch_max .or. alpha<1e-6) then
                !if (alpha < 0.01) then
                    ! par%step = par%step*0.5
                    goto 200
                endif
                goto 100
            endif        
            200 continue
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            if (nsearch > nsearch_max  .or.  alpha < 1e-6) then
            !if ( nsearch > nsearch_max ) then
                call update_slowness(par, npml, slow, step_min*d/100.0, slow_trial)
                slow = slow_trial
                c    = 1.0/slow
            else
                slow = 1.0/c
            endif
        ! + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + + !
        !!!! Qudratic line search
        else ! ( line_search ==1 )
            x2       = par%step
            y3       = -1.0
            nsearch  = 0
            step_min = x2
            res_min  = res_new
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            300 continue
            call update_slowness(par, npml, slow, x2*d/100.0, slow_trial)
            c    = 1.0/slow_trial
            call modeling_mpi( '2nd_order', .false., par, coord, s, c, den, fs, nx_pml, nz_pml, &
                                npml, damp_global, .true., y2)            
            if (rank == 0) then
                write(*,*) '   x2 = ', x2, ', res = ', y2
                write(1024,*) '   x2 = ', x2, ', res = ', y2
                flush(6)
                flush(1024)
            endif

            if (res_min > y2) then
                res_min = y2
                step_min = x2
            endif

            if ( y2 > res_new ) then
                x3 = x2
                y3 = y2
                x2 = x2/2.0
                nsearch = nsearch + 1                
                if ( nsearch> nsearch_max .or. x2<1.0e-6 ) then
                    par%step = par%step * 0.5
                    goto 400
                endif
                goto 300
            endif
            400 continue
            !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            if (y3 > 0.0) then
                alpha = quadratic_step(step_min,x3,res_new,res_min,y3)
                call  update_slowness( par, npml, slow, alpha*d/100.0, slow_trial )
                slow  = slow_trial
                c     = 1.0/slow
                call  modeling_mpi( '2nd_order', .false., par, coord, s, c, den, fs, nx_pml, nz_pml, &
                                    npml, damp_global, .true., res_min)
            else
                alpha = step_min
                call  update_slowness( par, npml, slow, alpha*d/100.0, slow_trial )
                slow  = slow_trial
                c     = 1.0/slow
            endif
            !call update_slowness( par, npml, slow, alpha*d/100.0, slow_trial )
            !slow = slow_trial
            !c    = 1.0/slow
            !call modeling_mpi( '2nd_order', .false., par, coord, s, c, den, fs, nx_pml, nz_pml, &
            !                    npml, damp_global, .true., res )
            if (rank == 0) then
                write(*, *) ' alpha = ', alpha, ', res = ', res_min
                write(1024,*) ' alpha = ', alpha, ', res = ', res_min
                flush(6)
                flush(1024)
            endif
        endif
        !---------------------------------------------------------------!
        ! END of step length determination
        !---------------------------------------------------------------!
        !my step length 
        if( ifmysteplen ) then
            ds  = 0.0
            res = 0.0
            do is=is1,is2
                boundary = 0.0
                call a2d42_modeling_is(is, par, coord, s, 1.0/(slow-dd)-c, fs, nx_pml, nz_pml, npml,&
                                       damp_global, .true., boundary, p_end, p_end_minus1, .true., res)
                !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                if (par%fileformat(1:2) == 'su') then
                    call filename(output, par%csg_out, is, '.su')
                    call read_sufile(output, csg_calc, par%nt, coord%ng(is), d1, d2)
                else
                    call filename(output, par%csg_out, is, '.bin')
                    call read_binfile(output, csg_calc, par%nt, coord%ng(is))
                endif
                !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                if (par%normalize == 1) then
                    obs_fac = max_abs(csg_obs, par%nt, coord%ng(is))
                    if (abs(obs_fac) < 1.0E-10) obs_fac = 1.0
                    csg_obs = csg_obs/obs_fac
    
                    calc_fac = max_abs(csg_calc, par%nt, coord%ng(is))
                    if (abs(calc_fac) < 1.0E-10) calc_fac = 1.0
                    csg_calc = csg_calc/calc_fac
                endif
    
                p0_bk = 0.0
                p1_bk = 0.0
                p_bk  = 0.0
                do ig=1,coord%ng(is),1+par%skiptrace
                    igx = npml + int(coord%xg(is,ig)/par%dx)+1
                    igz = npml + int(coord%zg(is,ig)/par%dx)+1
                    if (igz == fs(igx)) igz = igz+1
                    p0_bk(igz,igx) = dp(par%nt,ig)
                    p1_bk(igz,igx) = dp(par%nt-1,ig)
                enddo
    
                do it=par%nt-2,1,-1
                    call a2d42_recon_it( is,it, par, coord, s,c(npml+1:nz_pml-npml,npml+1:nx_pml-npml), &
                                         fs, boundary, p_end, p_end_minus1, p_recon )
     
                    call a2d42_back_it( is, it, par, coord, csg_calc, c, fs, nx_pml, nz_pml, npml, &
                                         damp_global, p0_bk, p1_bk, p_bk )
                    ds = ds + p_recon*p_bk(iz1:iz2,ix1:ix2)
                enddo
            enddo
            
            call MPI_Barrier(MPI_COMM_WORLD, ierr)
            call MPI_Allreduce(ds, g, par%nz*par%nx, MPI_REAL, MPI_SUM, MPI_COMM_WORLD, ierr)
            g = 2.0 * g * slow
            !step_min = sum(slow*slow)/sum(slow*g)
            write(*,*) 'calc_step_length = ', sum(slow*slow)/sum(slow*g)
        endif
        !-----------------------------------------------------------------------------------------!
        if (rank == 0 ) then 
            write(*, *) ' Chaiwoot''s step length = ', step_min
            write(*, *) '------------------------ Iteration ', iter, ' ends ---------------------------'
        endif
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
        ! Output velocity
        if (rank == 0) then
            call filename(output, 'vel-', iter, '.bin')
            !call write_sufile(output, c(iz1:iz2,ix1:ix2), par%nz, par%nx, par%dx, par%dx)
            call write_binfile( output, c(iz1:iz2,ix1:ix2), par%nz, par%nx )
        endif
        ! Update the old gradient
        g_old = g
    enddo  ! End of inversion loop
    !-------------------------------------------------------------------!
    ! Close the log file
    if ( rank == 0 ) then
        close(1024)
        close(1025)
    endif    
    close(7)
    deallocate( s, temp_src, c, ds,g,g_temp,d,dd,fs,damp,damp_global,p_end,p_end_minus1,&
                p_recon,csg_obs,csg_calc,dwin, dp,wvfield )   
    999 continue
    
    call stop_mpi
end subroutine slowness_inversion_2nd_order

!-----------------------------------------------------------------------------------------

subroutine slowness_inversion_staggered(parfile)

use datatype
use io
use math
use modeling
use mmi_mpi
use pml
use source

character(len=*), intent(in) :: parfile

! Show detailed messages during inversion or not
message = .true.

call start_mpi

! Read in velocity model parameters: nx, nz, dx, filename
call readparamfile(parfile,par)

! PML setting
call init_pml(par%nx, par%nz, par%npml)

! Read acquisition geometry data
call readcoordfile(par%coordfile,coord)

! Memory allocations
allocate(fs(nx_pml))
allocate(c(nz_pml,nx_pml))
allocate(slow(nz_pml,nx_pml))
allocate(slow_trial(nz_pml,nx_pml))
allocate(ds(par%nz,par%nx))
allocate(g(par%nz,par%nx))
allocate(g_old(par%nz,par%nx))
allocate(d(par%nz,par%nx))
allocate(boundary(par%nx*9+par%nz*6,par%nt))
allocate(p_end(par%nz,par%nx))
allocate(u_end(par%nz,par%nx))
allocate(w_end(par%nz,par%nx))
allocate(p0_bk(nz_pml,nx_pml))
allocate(p_bk(nz_pml,nx_pml))
allocate(u_bk(nz_pml,nx_pml))
allocate(w_bk(nz_pml,nx_pml))
allocate(csg_obs(par%nt,coord%ngmax))
allocate(csg_calc(par%nt,coord%ngmax))
allocate(dp(par%nt,coord%ngmax))
allocate(izwb(par%nx))

if (par%ismarine == 1) then
  if (par%waterbottomfile(1:3) == 'n/a') then
    izwb = par%izwb
  else
    ! Read waterbottom file
    call readwaterbottomfile(par, izwb)
  endif
endif

! Set up free surface
call free_surface(par, fs, npml)

! Read velocity model
call readvelfile(par,c,npml,nx_pml,nz_pml)
slow = 1.0/c

! Density setup
allocate(den(nz_pml,nx_pml))
if (par%densityfile(1:3) == 'n/a') then
  if (par%ismarine == 1) then
    call compute_density_wb(par%variable_density, c, den, nx_pml, nz_pml, npml, izwb)
  else
    call compute_density(par%variable_density, c, den, nx_pml, nz_pml)
  endif
else
  call read_densityfile(par,c,den,npml,nx_pml,nz_pml)
endif
if (rank == 0) call write_sufile('density.su',den,nz_pml,nx_pml,1.0,1.0)

! Determine minimum velocity
par%cmin = min_value(c(iz1:iz2,ix1:ix2),par%nz,par%nx,fs-npml)

! Setup PML damping coefficient
call setup_pml(par%dx, par%cmin)

! Setup Ricker source wavelet
allocate(s(par%nt))
call getsource(par, s)
if (rank == 0) call write_sufile('source.su',s,par%nt,par%dt)

! Log file
if (rank == 0) then
  write(*,*) 'Minimum velocity = ', par%cmin
  flush(6)
  open(20,file=par%logfile,form='formatted',status='replace')
  open(30,file=par%residualfile,form='formatted',status='replace')
endif

call get_assigned(1, coord%ns, is1, is2)

! Waveform inversion iteration
nsearch_max = 10
do iter=1,par%itermax

  ! Initialize the slowness update
  ds = 0.0
  res = 0.0

  do is=is1,is2,par%skipshot+1

    boundary = 0.0
    call staggered42_modeling_is(is, par, coord, s, c, den, fs, nx_pml, nz_pml, npml, damp_global, &
      .true., boundary, p_end, u_end, w_end, .true., res)

    if (par%fileformat(1:2) == 'su') then
      call filename(output, par%csg_in, is, '.su')
    else
      call filename(output, par%csg_in, is, '.bin')
    endif
    call read_sufile(output, csg_obs, par%nt, coord%ng(is), d1, d2)

    if (par%fileformat(1:2) == 'su') then
      call filename(output, par%csg_out, is, '.su')
    else
      call filename(output, par%csg_out, is, '.bin')
    endif
    call read_sufile(output, csg_calc, par%nt, coord%ng(is), d1, d2)

    if (par%normalize == 1) then
      ! Normalize the observed and calculated wavefields
      obs_fac = max_abs(csg_obs, par%nt, coord%ng(is))
      if (abs(obs_fac) < 1.0E-10) obs_fac = 1.0
      csg_obs = csg_obs/obs_fac

      calc_fac = max_abs(csg_calc, par%nt, coord%ng(is))
      if (abs(calc_fac) < 1.0E-10) calc_fac = 1.0
      csg_calc = csg_calc/calc_fac
    endif

    dp = csg_obs - csg_calc

    p_bk = 0.0
    u_bk = 0.0
    w_bk = 0.0

    do it=par%nt,2,-1
      p0_bk = p_bk
      call staggered42_recon_it(is, it, par, coord, s, c(iz1:iz2,ix1:ix2), den(iz1:iz2,ix1:ix2), &
        fs, boundary, p_end, u_end, w_end)

      call staggered42_back_it(is, it, par, coord, dp, c, den, fs, nx_pml, nz_pml, npml, damp_global, &
        p_bk, u_bk, w_bk)
      !ds = ds + p_end*p_bk(iz1:iz2,ix1:ix2)
      ds = ds + p_end*(p0_bk(iz1:iz2,ix1:ix2)-p_bk(iz1:iz2,ix1:ix2))
    enddo
  enddo

  call MPI_Barrier(MPI_COMM_WORLD, ierr)
  call MPI_Allreduce(ds, g, par%nz*par%nx, MPI_REAL, MPI_SUM, MPI_COMM_WORLD, ierr)
  call MPI_Allreduce(res, res_new, 1, MPI_REAL, MPI_SUM, MPI_COMM_WORLD, ierr)
  res_new = sqrt(res_new)

  ! Mute the gradient above the topography
  do ix=1,par%nx
    if (fs(ix+npml)-npml > 1) then
      g(1:fs(ix+npml)-npml,ix) = 0.0
    endif
  enddo

  ! Mute the gradient using the water bottom depth
  if (par%ismarine == 1) then
    call mute2(g,par%nx,par%nz,izwb,5)
  endif

  ! Compute the gradient wrt slowness
  g = 2.0*g*max_value(slow(iz1:iz2,ix1:ix2),par%nz,par%nx)/mean_abs(g,par%nz,par%nx)
  if (par%data(1:8) /= 'pressure') g = -g

  if (par%method == 1) then
    ! Conjugate Gradient
    if (iter == 1 .or. mod(iter,10) == 1) then
      d = g
    else
      beta = dot(g-g_old,g,par%nz,par%nx)/dot(g_old,g_old,par%nz,par%nx)
      d = g + beta * d
    endif
  else
    ! Steepest Descent
    d = g
  endif

  if (rank == 0) then
    write(*,*) 'iter ', iter, ', residual = ', res_new
    write(20,*) 'iter ', iter, ', residual = ', res_new
    write(30,*) iter, res_new
    flush(6)
    flush(20)
    flush(30)

    ! Output waveform gradient
    call write_sufile('grad.su',d,par%nz,par%nx,1.0,1.0)
  endif

  !----------------------------------------------------------
  ! Determine a step length using back-tracking line search
  !----------------------------------------------------------
  if (par%line_search == 0) then
    alpha = par%step
    nsearch = 0
    step_min = alpha
    res_min = res_new
    100 continue
    call update_slowness(par, npml, slow, alpha*d/100.0, slow_trial)
    c = 1.0/slow_trial
    call modeling_mpi('staggered', .false., par, coord, s, c, den, fs, nx_pml, nz_pml, &
                    npml, damp_global, .true., y2)
    if (rank == 0) then
      write(*,*) '   alpha = ', alpha, ', res = ', y2
      write(20,*) '   alpha = ', alpha, ', res = ', y2
      flush(6)
      flush(20)
    endif

    if (y2 > 0.95*res_new) then
      if (res_min > y2) then
        res_min = y2
        step_min = alpha
      endif
      alpha = alpha/2.0
      nsearch = nsearch+1
      if (nsearch > nsearch_max) then
        par%step = par%step*0.5
        goto 200
      endif
      goto 100
    endif
    200 continue
    if (nsearch > nsearch_max) then
      call update_slowness(par, npml, slow, step_min*d/100.0, slow_trial)
      slow = slow_trial
      c = 1.0/slow
    else
      slow = 1.0/c
    endif
  else
  !----------------------------------------------------------
  ! Determine a step length using quadratic line search
  !----------------------------------------------------------
    x2 = par%step
    y3 = -1.0
    nsearch = 0
    step_min = x2
    res_min = res_new
    300 continue
    call update_slowness(par, npml, slow, x2*d/100.0, slow_trial)
    c = 1.0/slow_trial
    call modeling_mpi('staggered', .false., par, coord, s, c, den, fs, nx_pml, nz_pml, &
                    npml, damp_global, .true., y2)
    if (rank == 0) then
      write(*,*) '   x2 = ', x2, ', res = ', y2
      write(20,*) '   x2 = ', x2, ', res = ', y2
      flush(6)
      flush(20)
    endif

    if (res_min > y2) then
      res_min = y2
      step_min = x2
    endif

    if (y2 > res_new) then
      x3 = x2
      y3 = y2
      x2 = x2/2.0
      nsearch = nsearch+1
      !if (nsearch > nsearch_max) then
      if (x2 < 1.0e-5) then
        par%step = par%step*0.5
        goto 400
      endif
      goto 300
    endif
    400 continue
    if (y3 > 0.0) then
      alpha = quadratic_step(step_min,x3,res_new,res_min,y3)
    else
      alpha = step_min
    endif
    call update_slowness(par, npml, slow, alpha*d/100.0, slow_trial)
    slow = slow_trial
    c = 1.0/slow
    call modeling_mpi('staggered', .false., par, coord, s, c, den, fs, nx_pml, nz_pml, &
                    npml, damp_global, .true., res)
    if (rank == 0) then
      write(*,*) '   alpha = ', alpha, ', res = ', res
      write(20,*) '   alpha = ', alpha, ', res = ', res
      flush(6)
      flush(20)
    endif
  endif
  !---------------------------------
  ! End of step length determination
  !---------------------------------

  ! Output velocity
  if (rank == 0) then
    call filename(output, par%velfile_out, iter, '.su')
    call write_sufile(output,c(iz1:iz2,ix1:ix2),par%nz,par%nx,par%dx,par%dx)
  endif

  ! Update the gradient for conjugate gradient method
  if (par%method == 1) then
    g_old = g
  endif
enddo  ! End of inversion loop

! Close the log file
if (rank == 0) then
  close(20)
  close(30)
endif

deallocate(s,c,den,ds,g,g_old,d,fs,damp,damp_global,p_end,u_end,w_end, &
           p_bk,u_bk,w_bk,csg_obs,csg_calc,dp)

999 continue

call MPI_Finalize(ierr)

end subroutine slowness_inversion_staggered

end module apps2d

